

#include <stdlib.h>
#include "osdobjects.h"
#include "osdserver.h"

// --------------------
//   cOsdServerObject
// --------------------

cMutex cShadowBase::LockShared;

cOsdServerObject::cOsdServerObject() {
    name = NULL;
    context = NULL;
    id=0;
}


cOsdServerObject::~cOsdServerObject() {
    if (context) context->Del(this,false);
    if (name) free(name);
}


cOsdServerObject* cServerContext::GetName(const char *Name) {
    if (!Name) return NULL;
    for (cOsdServerObject *obj = First() ; obj ; obj = Next(obj)) {
        const char *objname=obj->Name();
        if (!objname) continue;
        if (0 == strcasecmp(Name,objname)) return obj;
    }
    return NULL;
}


// ---------------------
//   cOsdServerContext
// ---------------------


void cServerContext::Add(cOsdServerObject *Object, const char *Name) {
    maxid++;
    Object->id = maxid;
    if (Name) {
        cOsdServerObject *prev=GetName(Name);
        if (prev) {
            free(prev->name);
            asprintf(&prev->name,"_%i",prev->id);
        }
        Object->name = strdup(Name);
    } else {
        asprintf(&Object->name,"_%i",Object->id);
    }
    Object->context = this;
    cList<cOsdServerObject>::Add(Object);
}

cServerContext::~cServerContext() {
    // Avoid the default clear destructor - objects removes itself from list

    while (First()) delete First();
}




// ---------------
//   cShadowBase
// ---------------


cShadowBase::cShadowBase() {
    Dirty=true;
}

cShadowBase::cPrivateBase::cPrivateBase(cShadowBase *parent) {
    Parent=parent;
}

cShadowBase::cPrivateBase::~cPrivateBase() {
    if (!Parent) return;
    esyslog("osdserver: hard detaching of OSD VDR object");
    Parent->Detach();
}

// -----------------------
//   cShadowTemplate<T>
// -----------------------


template<class T,class B> cShadowTemplate<T,B>::cShadowTemplate() {
    Private=NULL;
}

template<class T,class B> cShadowTemplate<T,B>::~cShadowTemplate() {
    cMutexLock Lock(&B::LockShared);
    if (Private) {
        esyslog("osdserver: hard detaching of OSD shadow object");
        Detach();
    }
}

template<class T,class B> void cShadowTemplate<T,B>::Detach() {
    // Detach private VDR object, make it independent
    cMutexLock Lock(&B::LockShared);

    if (!Private) return; // already detached

    Private->Parent=NULL;
    Private=NULL;
}



// ---------------------
//   cOsdServerOsdItem
// ---------------------


cOsdServerOsdItem::cOsdServerOsdItem(const char *Text, bool Selectable) {
    text=strdup(Text);
    selectable=Selectable;
}

cOsdServerOsdItem::~cOsdServerOsdItem() {
    Detach();
    if (text) free(text);
}

cOsdServerOsdItem::cPrivateOsdItem::cPrivateOsdItem(cOsdServerOsdItem *parent)
    : cPrivate(parent) {
    cOsdItem::SetText(parent->text);
    cOsdItem::SetSelectable(parent->selectable);
}

eOSState cOsdServerOsdItem::cPrivateOsdItem::ProcessKey(eKeys Key) {
/*
    if (Key==kOk) {
        cMutexLock Lock(&LockShared);
        if (ParentOsdItem() && !ParentOsdItem()->Dirty) ParentOsdItem()->OkEvent=true;

        return osEnd;
    }
*/
    return osUnknown;
}

void cOsdServerOsdItem::Update() {
    if (!Private) {
        Private=new cPrivateOsdItem(this);
    } else if (Dirty) {
        PrivateOsdItem()->SetText(text);
        PrivateOsdItem()->SetSelectable(selectable);
    }
    Dirty=false;
}

bool cOsdServerOsdItem::PollEvent(cOsdServerObject::cEvent &Event) {
/*
    cMutexLock Lock(&LockShared);
    if (OkEvent) {
        Event.ev=cEvent::OkEvent;
        Event.who=this;
        OkEvent=false;
        return true;
    }
*/
    return false;
}

// ---------------------
//   cOsdServerOsdMenu
// ---------------------

cOsdServerMenu::cOsdServerMenu(const char *Title) {
    title=strdup(Title);
    current=0;
    CloseEvent=false;
    OkEvent=false;
    CloseEventEnabled=false;
    OkEventEnabled=false;
    SendState=osUnknown;
}

cOsdServerMenu::~cOsdServerMenu() {
    Detach();
    if (title) free(title);
}

cOsdServerMenu::cPrivateMenu::cPrivateMenu(cOsdServerMenu *parent)
    : cPrivate(parent) {
    SetTitle(parent->title);
}

cOsdServerMenu::cPrivateMenu::~cPrivateMenu() {
    cMutexLock Lock(&LockShared);

    if (ParentMenu() && !ParentMenu()->Dirty && ParentMenu()->CloseEventEnabled) ParentMenu()->CloseEvent=true;

    if (Parent) Parent->Detach();
}


eOSState cOsdServerMenu::cPrivateMenu::ProcessKey(eKeys Key) {
    if (!Parent) return osEnd;
    // We're detached, close this menu

    eOSState state = cOsdMenu::ProcessKey(Key);

    cMutexLock Lock(&LockShared);

    if (ParentMenu() && !ParentMenu()->Dirty) ParentMenu()->current=Current();

    if (ParentMenu()->SendState != osUnknown) {
        state=ParentMenu()->SendState;
        ParentMenu()->SendState=osUnknown;
        return state;
    }

    if (Key==kOk) {
        if (ParentMenu() && !ParentMenu()->Dirty && ParentMenu()->OkEventEnabled) ParentMenu()->OkEvent=true;
    }

    return state;
}

void cOsdServerMenu::Update() {
    if (!Private) {
        Private=new cPrivateMenu(this);
    } else if (Dirty) {
        PrivateMenu()->SetTitle(title);
    }

    // Now call Update for all menu items:
    cOsdItem *last=NULL;
    for (cOsdServerMenuItem *item = menuitems.First(); item; item=menuitems.Next(item)) {
        item->Update();

        // Check ordering of VDR menu items
        cOsdItem *osditem=item->GetItem();
        if (osditem->Prev() == NULL && osditem != PrivateMenu()->First()) {
            // Not yet listed, probably new
            PrivateMenu()->Add(osditem,false,last);
        } else if (osditem->Prev() != last) {
            // Not listed correctly, move it
            cOsdMenu *menu=PrivateMenu();
            menu->cList<cOsdItem>::Del(osditem,false);
            PrivateMenu()->Add(osditem,false,last);
        }
        last=osditem;
    }

    // TODO: Check for deleted menu items.
    // They are now after last in the list

    if (Dirty) {
        cOsdServerMenuItem *cur=menuitems.Get(current);
        if (cur) PrivateMenu()->SetCurrent(cur->GetItem());
        current=PrivateMenu()->Current(); // should be the same, just to make sure
    }

    Dirty=false;
}

void cOsdServerMenu::Detach() {
    // Call Detach for all menu items too
    cMutexLock Lock(&LockShared);
    for (cOsdServerMenuItem *item = menuitems.First(); item; item=menuitems.Next(item)) {
        item->Detach();
    }
    cShadowTemplate<cOsdMenu_,cShadowBase>::Detach();
}


void cOsdServerMenu::Add(cOsdServerMenuItem *Item, bool Current, cOsdServerMenuItem *After) {
    cMutexLock Lock(&LockShared);
    Item->menu=this;
    menuitems.Add(Item,After);
    if (Current) current=Item->Index();
    Dirty=true;
}

void cOsdServerMenu::Ins(cOsdServerMenuItem *Item, bool Current, cOsdServerMenuItem *Before) {
    cMutexLock Lock(&LockShared);
    Item->menu=this;
    menuitems.Ins(Item,Before);
    if (Current) current=Item->Index();
    Dirty=true;
}

cOsdObject* cOsdServerMenu::OpenOsd() {
    Update();
    return PrivateMenu();
}

void cOsdServerMenu::Show() {
    cMainThreadLock Lock;
    Lock.LockMainThread();

    if (PrivateMenu()) {
        // There's already a VDR side object, just call show
        Update();
        PrivateMenu()->Show();

        Lock.UnlockMainThread();
        return;
    }

    Lock.UnlockMainThread();

    // Need to open the menu first. Call callback to do it.
    cPluginOsdServer::Plugin()->CallMainMenuAction(this,5000);
}


void cOsdServerMenu::GetCurrent(int &cur, const char *&curs) {
    cur=current;
    cOsdServerMenuItem *o=menuitems.Get(cur);
    curs=o ? o->Name() : NULL;
}

void cOsdServerMenu::SetCurrent(int cur) {
    cMutexLock Lock(&LockShared);
    current=cur;
    Dirty=true;
}

void cOsdServerMenu::SendOSState(eOSState State) {
    cMutexLock Lock(&LockShared);

    SendState=State;
    cRemote::Put(eKeys(kNone | k_Release));
}

void cOsdServerMenu::EnableKeyEvent(int key) {
    cMutexLock Lock(&LockShared);

    if (key==kOk) OkEventEnabled=true;
}

void cOsdServerMenu::EnableCloseEvent() {
    cMutexLock Lock(&LockShared);
    CloseEventEnabled=true;
}


bool cOsdServerMenu::PollEvent(cEvent &Event) {
    cMutexLock Lock(&LockShared);
    for (cOsdServerMenuItem *item = menuitems.First(); item; item=menuitems.Next(item)) {
        if (item->PollEvent1(Event)) return true;
    }
    if (CloseEventEnabled && CloseEvent) {
        Event.ev=cEvent::CloseEvent;
        Event.who=this;
        CloseEvent=false;
        return true;
    }
    if (OkEventEnabled && OkEvent) {
        Event.ev=cEvent::OkEvent;
        Event.who=this;
        OkEvent=false;
        return true;
    }
    return false;
}
