#ifndef __OSDSERVER_OSDOBJECTS_H
#define __OSDSERVER_OSDOBJECTS_H

#include <vdr/tools.h>
#include <vdr/thread.h>
#include <vdr/osdbase.h>

#include "tools.h"

class cServerContext;


class cOsdServerObject : public cListObject {
    // Generic base class for named and unnamed objects of the network interface.
    // These objects live in the network thread
    // These objects may 'own' a VDR main thread object.
    // The OsdServerObject holds a copy of the state of the VDR object, and
    // can push its state to the VDR object, or create it if it doesnt exist yet.
    // The VDR object can detach if its lifetime ends.
    // Before deleting this object, always call Detach().

private:
    char *name;
    int id;
    cServerContext *context;

public:
    cOsdServerObject();
    virtual ~cOsdServerObject();

    // IDs to identify certain object instances from base class
    enum enumClassId {
        clsidMenu,
        clsidOsdItem
    };
    virtual enumClassId ClassId()=0;

    // Generic cast function that checks the id.
    // Use as baseobject->Cast<type>()
    template<class T> T* Cast() {
        if (ClassId() != T::classid) return NULL;
        return dynamic_cast<T*>(this);
    }

    // Context object name or NULL if not named
    const char* Name() { return name; }


    class cEvent {
    public:
        enum { NoEvent, OkEvent, CloseEvent } ev;
        cOsdServerObject *who;

        cEvent() { ev=NoEvent; who=NULL; }
    };

    virtual bool PollEvent(cEvent &Event) { return false; }
    // Primitive event handling

    friend class cServerContext;
};



class cServerContext : private cList<cOsdServerObject> {
    // Manage a collection of named OsdServerObjects.

    int maxid;

public:
    cOsdServerObject* GetName(const char *Name);
    // Get object by name

    void Add(cOsdServerObject *Object, const char *Name);
    // Also sets the name of the object
    // TODO: Duplicate name detection

    virtual ~cServerContext();

    friend class cOsdServerObject;
};





class cShadowBase {
    // Abstract base class for shadowed objects. Acts as base class for cShadowTemplate

protected:
    bool Dirty;
    // =true means that this object has changes that are not yet Update()d to
    // the VDR object. Also means that the VDR object will not upload
    // local changes, they will be discarded on Update().
    // Protected by LockShared.

    static cMutex LockShared;
    // Common lock to access certain members from the VDR main thread

    class cPrivateBase {
    public:
        cShadowBase *Parent;
        // Parent is a pointer to the shadow object, or NULL if detached.
        // This member is protected by LockShared mutex.

        cPrivateBase(cShadowBase *parent);
        // Constructors are always called with synced threads

        virtual ~cPrivateBase();
        // Warning: Higher level destructors need to call Parent->Detach().
        // Foreground thread always calls destructor on its own
    };

public:
    virtual void Update()=0;
    // Push object status to VDR object or create VDR object.
    // Both threads MUST be synchronized before calling this

    virtual void Detach()=0;
    // Detach VDR object from this object. Can be called
    // from any context.

    virtual bool IsDetached()=0;
    // Check if object is 'live'

    cShadowBase();
    virtual ~cShadowBase() {};
};


class cOsdServerMenu;

class cOsdServerMenuItem : public cListObject, public cShadowBase  {
    // Base class for menu items
    cOsdServerMenu *menu;
protected:
    virtual cOsdItem* GetItem()=0;

    cOsdServerMenuItem() { menu=NULL; }

public:
    cOsdServerMenu* GetMenu() { return menu; }
    virtual int GetIndex()=0;
    virtual const char* Name()=0;

    virtual bool PollEvent1(cOsdServerObject::cEvent &Event)=0;
    // Primitive event handling

    friend class cOsdServerMenu;
};




template<class T, class B> class cShadowTemplate : public B {
    // VDR object shadow class
    // This template encapsulates a VDR object as cOsdServerShadow<T>::cPrivate
    // and provides the shadow cOsdServerObject for it.
    // cPrivate lives in the VDR main thread, while cOsdServerShadow lives in the network thread.
    // T needs to have a parameter-less constructor. Make one if you need one.

protected:

    class cPrivate : public T, public B::cPrivateBase {
        public:
        cPrivate(cShadowTemplate<T,B> *parent) : B::cPrivateBase(parent) { };

        friend class cShadowTemplate<T,B>;
    };

    cPrivate *Private;
    // Pointer to the shadowed VDR object, or NULL if none attached (yet).
    // Call Upade() to attach one.

public:
    cShadowTemplate<T,B>();
    virtual ~cShadowTemplate<T,B>();
    // Constructor and destructor are called from network thread.
    // Warning: Higher level destructors need to call Detach().

    virtual void Detach();
    // This Detach() removes the cross-links between cOsdServerShadow and cPrivate.

    virtual bool IsDetached() { return Private==NULL; }
};

template<class T> class cShadowObjectTemplate : public cShadowTemplate<T,cShadowBase> {
};

template<class T> class cShadowMenuItemTemplate : public cOsdServerObject, public cShadowTemplate<T,cOsdServerMenuItem> {
public:
    virtual cOsdItem* GetItem() { return cShadowTemplate<T,cOsdServerMenuItem>::Private; }
    virtual int GetIndex() { return cOsdServerMenuItem::Index(); }
    virtual const char* Name() { return cOsdServerObject::Name(); }
    virtual bool PollEvent1(cOsdServerObject::cEvent &Event) { return PollEvent(Event); }
};




class cOsdServerOsdItem : public cShadowMenuItemTemplate<cOsdItem> {
    // Shadow of cOsdItem

    // Shadow variables of cOsdItem
    char *text;
    bool selectable;
    // These are protected by LockShared mutex,
    // unless both threads are synchronized anyway.

    class cPrivateOsdItem : public cPrivate {
        public:
        cPrivateOsdItem(cOsdServerOsdItem *parent);
        virtual ~cPrivateOsdItem() { if (Parent) Parent->Detach(); }
        virtual eOSState ProcessKey(eKeys Key);
        cOsdServerOsdItem* ParentOsdItem() { return (cOsdServerOsdItem*)Parent; }
    };
    cPrivateOsdItem* PrivateOsdItem() { return (cPrivateOsdItem*)Private; }

public:
    cOsdServerOsdItem(const char *Text, bool Selectable = true);
    virtual ~cOsdServerOsdItem();

    virtual void Update();

    virtual bool PollEvent(cEvent &Event);

    static const enumClassId classid=clsidOsdItem;
    virtual enumClassId ClassId() { return classid; }
};



class cOsdMenu_ : public cOsdMenu {
    // Wrapper to add constructor
    public: cOsdMenu_() : cOsdMenu("") { }
};

class cOsdServerMenu : public cOsdServerObject, public cShadowObjectTemplate<cOsdMenu_>, public cOpenOsd::cCallback {
    // Shadow of cOsdMenu

    // Shadow variables of cOsdMenu
    char *title;
    cList<cOsdServerMenuItem> menuitems;
    int current;

    eOSState SendState;

    bool CloseEvent;
    bool OkEvent;
    bool CloseEventEnabled;
    bool OkEventEnabled;
    // These are protected by LockShared mutex,
    // unless both threads are synchronized anyway.

    class cPrivateMenu : public cPrivate {
        public:
        cPrivateMenu(cOsdServerMenu *parent);
        virtual ~cPrivateMenu();
        cOsdServerMenu* ParentMenu() { return (cOsdServerMenu*)Parent; }

        virtual eOSState ProcessKey(eKeys Key);
        // Closes menu if it was detached

        // Make some protected stuff public
        void SetTitle(const char *title) { cOsdMenu::SetTitle(title); }
        void SetCurrent(cOsdItem *Item) { cOsdMenu::SetCurrent(Item); }
    };
    cPrivateMenu* PrivateMenu() { return (cPrivateMenu*)Private; }

public:
    cOsdServerMenu(const char *Title);
    virtual ~cOsdServerMenu();

    virtual void Update();
    virtual void Detach();

    void Add(cOsdServerMenuItem *Item, bool Current = false, cOsdServerMenuItem *After = NULL);
    void Ins(cOsdServerMenuItem *Item, bool Current = false, cOsdServerMenuItem *Before = NULL);

    virtual cOsdObject* OpenOsd();
    // Callback for opening the OSD from main thread

    void Show();
    // Calls cOsdMenu::Show() or creates the cOsdMenu object and displays it. Does the necessary thread syncing.

    void GetCurrent(int &cur, const char *&curs);
    void SetCurrent(int cur);

    void SendOSState(eOSState State);

    void EnableKeyEvent(int key);
    void EnableCloseEvent();

    virtual bool PollEvent(cEvent &Event);

    static const enumClassId classid=clsidMenu;
    virtual enumClassId ClassId() { return classid; }
};





#endif
